#include <stdint.h>
#include <avr/io.h>

#include "i2c.h"

// Initialise (and enable) the two-wire interface with a particular clock frequency.
void i2c_init(uint32_t frequency) {
	
	// Set TWPS to= 0
	TWSR &= (uint8_t)~(_BV(TWPS1) | _BV(TWPS0));
	// Set the SCL frequency [ F_CPU / (16 + 2 * TWBR * (4 ** TWPS)) ].
	TWBR = ((F_CPU / frequency) - 16) / 2;
	
	// Enable two-wire interface.
	TWCR = _BV(TWEN);
}

// Write a block of data to a two-wire device.
uint8_t i2c_write(uint8_t slave_address, uint8_t target_address, uint8_t* data, uint8_t length) {
	
	// Send START.
	TWCR = _BV(TWINT) | _BV(TWSTA) | _BV(TWEN);
	while (!(TWCR & _BV(TWINT)));
	if ((TWSR & I2C_STATUS_MASK) != I2C_STATUS_START) return I2C_STATUS_START;
	
	// Send slave address.
	TWDR = slave_address & 0xFE;
	TWCR = _BV(TWINT) | _BV(TWEN);
	while (!(TWCR & _BV(TWINT)));
	if ((TWSR & I2C_STATUS_MASK) != I2C_STATUS_SLA_W_ACK) return I2C_STATUS_SLA_W_ACK;
	
	// Send target address.
	TWDR = target_address;
	TWCR = _BV(TWINT) | _BV(TWEN);
	while (!(TWCR & _BV(TWINT)));
	if ((TWSR & I2C_STATUS_MASK) != I2C_STATUS_DATA_W_ACK) return I2C_STATUS_DATA_W_ACK;
	
	// Send data.
	while (length--) {
		TWDR = *data++;
		TWCR = _BV(TWINT) | _BV(TWEN);
		while (!(TWCR & _BV(TWINT)));
		if ((TWSR & I2C_STATUS_MASK) != I2C_STATUS_DATA_W_ACK) return I2C_STATUS_DATA_W_ACK;
	}
	
	// Send STOP.
	TWCR = _BV(TWINT) | _BV(TWEN) | _BV(TWSTO);
	
	// Success!
	return 0;
}

// Read a block of data from a two-wire device.
uint8_t i2c_read(uint8_t slave_address, uint8_t source_address, uint8_t* data, uint8_t length) {

	// Send START.
	TWCR = _BV(TWINT) | _BV(TWSTA) | _BV(TWEN);
	while (!(TWCR & _BV(TWINT)));
	if ((TWSR & I2C_STATUS_MASK) != I2C_STATUS_START) return I2C_STATUS_START;
	
	// Send slave address.
	TWDR = slave_address & 0xFE;
	TWCR = _BV(TWINT) | _BV(TWEN);
	while (!(TWCR & _BV(TWINT)));
	if ((TWSR & I2C_STATUS_MASK) != I2C_STATUS_SLA_W_ACK) return I2C_STATUS_SLA_W_ACK;
	
	// Send target address.
	TWDR = source_address;
	TWCR = _BV(TWINT) | _BV(TWEN);
	while (!(TWCR & _BV(TWINT)));
	if ((TWSR & I2C_STATUS_MASK) != I2C_STATUS_DATA_W_ACK) return I2C_STATUS_DATA_W_ACK;

	// Send RESTART.
	TWCR = _BV(TWINT) | _BV(TWSTA) | _BV(TWEN);
	while (!(TWCR & _BV(TWINT)));
	if ((TWSR & I2C_STATUS_MASK) != I2C_STATUS_REPEATED_START) return I2C_STATUS_REPEATED_START;
	
	// Send slave address.
	TWDR = slave_address | 0x01;
	TWCR = _BV(TWINT) | _BV(TWEN);
	while (!(TWCR & _BV(TWINT)));
	if ((TWSR & I2C_STATUS_MASK) != I2C_STATUS_SLA_R_ACK) return I2C_STATUS_SLA_R_ACK;
	
	// Receive data.
	while (length--) {
		if (length) {
			TWCR = _BV(TWINT) | _BV(TWEN) | _BV(TWEA);
			while (!(TWCR & _BV(TWINT)));
			if ((TWSR & I2C_STATUS_MASK) != I2C_STATUS_DATA_R_ACK) return I2C_STATUS_DATA_R_ACK;
		} else {
			TWCR = _BV(TWINT) | _BV(TWEN);
			while (!(TWCR & _BV(TWINT)));
			if ((TWSR & I2C_STATUS_MASK) != I2C_STATUS_DATA_R_NOT_ACK) return I2C_STATUS_DATA_R_NOT_ACK;
		}
		*data++ = TWDR;
	}
	
	// Send STOP.
	TWCR = _BV(TWINT) | _BV(TWEN) | _BV(TWSTO);
	
	// Success!
	return 0;
}