#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>

#include <avr/interrupt.h>
#include <avr/io.h>
#include <avr/pgmspace.h>

#include <util/delay.h>

#include "spi.h"
#include "i2c.h"

#include "z80.h"
#include "z80_memory.h"

#include "sd.h"
#include "partitions.h"
#include "keyboard.h"
#include "serial.h"
#include "parallel.h"
#include "date_time.h"

#include "bcd.h"
#include "vdu.h"
#include "vgm.h"

// Writes a byte to the active console (as determined by the AVR).
void console_output(char c) {
	vdu_write(c);
	serial_transmit(c);
}

// Writes a NUL-terminated string to the VDU.
void console_output_string(char* s) {
	char c;
	while ((c = *s++)) {
		console_output((uint8_t)c);
	}
}

// Writes a NUL-terminated string stored in ROM to the console.
void console_output_string_P(const char* s) {
	char c;
	while ((c = pgm_read_byte(s++))) {
		console_output(c);
	}
}

void console_output_hex_nybble(uint8_t n) {
	n &= 0xF;
	if (n > 9) {
		console_output(n - 10 + 'A');
	} else {
		console_output(n + '0');
	}
}

void console_output_hex_byte(uint8_t b) {
	console_output_hex_nybble(b >> 4);
	console_output_hex_nybble(b);
}

void console_output_hex_word(uint16_t w) {
	console_output_hex_byte(w >> 8);
	console_output_hex_byte(w);
}

void console_output_hex_int(uint32_t i) {
	console_output_hex_word(i >> 16);
	console_output_hex_word(i);
}

ISR(BADISR_vect) {
	console_output_string_P(PSTR("\r\nUnhandled interrupt.\r\n"));
    for(;;);
}

ISR(PCINT3_vect) {
	switch (keyboard_transmission_direction) {
		case KEYBOARD_TRANSMISSION_DEVICE_TO_HOST:
			if (!(KEYBOARD_CLOCK_PIN & _BV(KEYBOARD_CLOCK))) {
				if (TIFR0 & _BV(TOV0)) keyboard_clear_receive_buffer();
				keyboard_bit_received(KEYBOARD_DATA_PIN & _BV(KEYBOARD_DATA));
				TCNT0 = 0;
				TIFR0 |= _BV(TOV0);
			}
			break;
		case KEYBOARD_TRANSMISSION_HOST_TO_DEVICE:
			keyboard_send_bit();
			break;
	}
}

void keyboard_disable_clock_interrupt() {
	PCICR &= (uint8_t)~_BV(PCIE3);
	PCMSK3 &= (uint8_t)~_BV(PCINT29);
}

void keyboard_enable_clock_interrupt() {
	PCICR |= _BV(PCIE3);
	PCMSK3 |= _BV(PCINT29);
}

void keyboard_wait_key(void) {
	while (keyboard_buffer_dequeue());
	while (keyboard_buffer_dequeue() == 0);
	while (keyboard_buffer_dequeue());
	keyboard_key_received_flag = 0;
}

void sd_read_record_notify_errors(uint32_t address, uint8_t* buffer) {
	while (sd_read_record(address, buffer) != SD_ERROR_OK) {
		console_output_string_P(PSTR("Error reading from SD card. Strike key to try again."));
		keyboard_wait_key();
		console_output_string_P(PSTR("\r\n"));
	}
}

void sd_read_block_notify_errors(uint32_t address, uint8_t* buffer) {
	while (sd_read_block(address, buffer) != SD_ERROR_OK) {
		console_output_string_P(PSTR("Error reading from SD card. Strike key to try again."));
		keyboard_wait_key();
		console_output_string_P(PSTR("\r\n"));
	}
}

int main(void) {

	// Brief initial delay to allow ICs to respond.
	_delay_ms(100);
	
	// Initialise hardware.
	z80_init();	
	serial_init(9600);
	parallel_init();
	i2c_init(100000UL);
	
	// There should be 100uS between incoming keyboard bits.
	// 100uS = 2000 clock cycles at 20MHz.
	// With a prescaler of 8, TIM0 would overflow after 8*256 = 102.4uS
	TCCR0B |= _BV(CS01); // clkIO/8 (From prescaler).
	keyboard_reset();
	
	// Set up timer 1 to run at 100Hz.
	TCCR1B |= (0b010 << CS10); // Run at clkIO/8.
	TCCR1B |= _BV(WGM12);      // Mode 4 (TOP=OCR1A, CTC).
	OCR1A = (F_CPU / 8 / 100) - 1;

	// Enable interrupts.
	sei();

	// Longer delay to give everything else time to settle.
	_delay_ms(250);
	
	console_output_string_P(PSTR("Starting...\r\n\r\n"));

	// Enable numlock.
	keyboard_byte_received(0x77); // Simulate pressing Num Lock.

	// Reset the SD card.
	while (sd_reset() != SD_ERROR_OK) {
		console_output_string_P(PSTR("SD card not responding. Strike key to try again."));
		keyboard_wait_key();
		console_output_string_P(PSTR("\r\n"));
	}
	
	// Fetch the CSD.
	struct sd_csd_t sd_csd;
	while (sd_read_csd(&sd_csd) != SD_ERROR_OK) {
		console_output_string_P(PSTR("SD card not reporting CSD. Strike key to try again."));
		keyboard_wait_key();
		console_output_string_P(PSTR("\r\n"));
	}		
	
	// Calculate the size of the card.
	uint16_t mult = 1 << (sd_csd.C_SIZE_MULT + 2);
	uint16_t block_len = 1 << sd_csd.READ_BL_LEN;
	uint32_t capacity = (sd_csd.C_SIZE + 1UL) * (uint32_t)mult * (uint32_t)block_len;
	sprintf((char*)sd_block, "%lu", capacity);
	
	// Display the size of the card.
	console_output_string_P(PSTR("SD card capacity: "));		
	console_output_string((char*)sd_block);
	console_output_string_P(PSTR(" bytes.\r\n\r\n"));
	
	// Read the partition table.
	sd_read_record_notify_errors(0, sd_block);		
	for (uint8_t partition = 0; partition < 16; ++partition) {
		partitions[partition] = *(struct partition_t*)(sd_block + partition * 32);
	}
	
	for (uint8_t partition = 0; partition < 16; ++partition) {
		if (partitions[partition].offset != 0) {
			sprintf((char*)sd_block, "Drive %c: %lu bytes.\r\n", 'A' + partition, partitions[partition].total_size);
			console_output_string((char*)sd_block);
		}
	}
	console_output_string_P(PSTR("\r\n"));
	
	struct partition_t* current_partition = partitions;

	
	for (;;) {
	
		// Switch back to drive A:
		current_partition = partitions;
		
		// Re-initialise I/O ports.
		serial_init(9600);
		parallel_init();
		
		z80_reset();
		
		z80_control_bus();
		
		// Read CPM3.SYS header from boot device.
		sd_read_record_notify_errors(current_partition->offset, sd_block);
		uint16_t resident_destination = sd_block[0] << 8;
		uint8_t resident_length = sd_block[1];
		//uint16_t banked_destination = sd_block[2] << 8;
		//uint8_t banked_length = sd_block[3];
		uint16_t cold_boot_entry_point = sd_block[4] | (sd_block[5] << 8);
		
		// Display loader message.
		console_output_string_P(PSTR("CP/M Version 3.0 Loader\r\n"));
		console_output_string((char*)(sd_block + 16));
		console_output_string_P(PSTR("\r\n"));

		// Display print record (load table).
		sd_read_record_notify_errors(current_partition->offset + 128, sd_block);
		{
			char* p = (char*)sd_block;
			char c;
			while ((c = *p++) != '$') console_output(c);
		}
		
		// Load CPM3.SYS.
		uint16_t z80_address = resident_destination - 128;
		uint32_t flash_address = current_partition->offset + 0x100;
		uint16_t record_count = resident_length * 2;
		
		while (record_count-- > 0) {
			
			// Read a record.
			sd_read_record_notify_errors(flash_address, sd_block);
			
			// Copy it to Z80 memory.
			uint16_t target_address = z80_address;
			z80_memory_control_address();
			for (uint8_t read_buffer = 0; read_buffer < 128; ++target_address, ++read_buffer) {
				z80_memory_set_address(target_address);
				z80_memory_write(sd_block[read_buffer]);
			}		
			z80_memory_release_address();
			
			// Advance pointers.
			flash_address += 128;
			z80_address -= 128;
		}
		
		z80_memory_control_address();
		
		uint16_t boot_stub_address = 0;		
		z80_memory_set_address(boot_stub_address++); z80_memory_write(0xC3); // JP
		z80_memory_set_address(boot_stub_address++); z80_memory_write(cold_boot_entry_point);
		z80_memory_set_address(boot_stub_address++); z80_memory_write(cold_boot_entry_point >> 8);
		
		z80_memory_release_address();
		
		z80_release_bus();
		
		z80_reset();
		
		struct date_time_t date_time;
		
		if (!date_time_rtc_is_valid()) {
			date_time.hour = 0x00;
			date_time.minute = 0x00;
			date_time.second = 0x00;
			date_time.date = 1;
			date_time_set_rtc(&date_time);
			console_output_string_P(PSTR("\r\nDate and time reset."));
		}
		
		uint16_t selected_dma = 0;
		uint16_t selected_sector = 0;
		uint16_t selected_track = 0;
		
		keyboard_reset_flag = 0;
		
		uint8_t interrupt_enable = 0;
		uint8_t interrupt_flags = 0;
		
		uint8_t inkey_code = 0;
		
		while (!keyboard_reset_flag) {
		
			// Have any interrupts occurred?
			if ((interrupt_enable & 0x01) && keyboard_key_received_flag) {
				keyboard_key_received_flag = 0;
				interrupt_flags |= 0x01;
			}
			if ((interrupt_enable & 0x04) && serial_received_flag) {
				serial_received_flag = 0;
				interrupt_flags |= 0x04;
			}
			if ((interrupt_enable & 0x40) && (TIFR1 & _BV(OCF1A))) {
				TIFR1 |= _BV(OCF1A);
				interrupt_flags |= 0x40;
			}
			if (interrupt_flags) {
				// Assert /INT
				Z80_INT_PORT &= (uint8_t)~_BV(Z80_INT);
			}
		
			// Handle /IORQ.
			if (!(Z80_IORQ_PIN & _BV(Z80_IORQ))) {
			
				if (Z80_M1_PIN & _BV(Z80_M1)) {
				
					uint8_t address = z80_memory_get_address();
					uint8_t wait_reset = 0;
					if (!(Z80_WR_PIN & _BV(Z80_WR))) {
						// Respond to Z80 write.
						switch (address) {
							case 0x00: // vdu
								vdu_write(Z80_DATA_PIN);
								break;
							case 0x06: // key$scan
								inkey_code = Z80_DATA_PIN;
								break;
							case 0x08: // serial$out
								serial_transmit(Z80_DATA_PIN);
								break;
							case 0x12: // disk$dma$l
								selected_dma = (selected_dma & 0xFF00) | Z80_DATA_PIN;
								break;
							case 0x13: // disk$dma$h
								selected_dma = (selected_dma & 0x00FF) | (Z80_DATA_PIN << 8);
								break;
							case 0x14: // disk$sector$l
								selected_sector = (selected_sector & 0xFF00) | Z80_DATA_PIN;
								break;
							case 0x15: // disk$sector$h
								selected_sector = (selected_sector & 0x00FF) | (Z80_DATA_PIN << 8);
								break;
							case 0x16: // disk$track$l
								selected_track = (selected_track & 0xFF00) | Z80_DATA_PIN;
								break;
							case 0x17: // disk$track$h
								selected_track = (selected_track & 0x00FF) | (Z80_DATA_PIN << 8);
								break;
							case 0x18: // disk$drive
								current_partition = &(partitions[Z80_DATA_PIN]);
								break;
							case 0x20: // time
								date_time_set_rtc(&date_time);
								break;
							case 0x21: // time$date$l
								date_time.date = (date_time.date & 0xFF00) | Z80_DATA_PIN;
								break;
							case 0x22: // time$date$h
								date_time.date = (date_time.date & 0x00FF) | (Z80_DATA_PIN << 8);
							case 0x23: // time$hour
								date_time.hour = Z80_DATA_PIN;
								break;
							case 0x24: // time$minute
								date_time.minute= Z80_DATA_PIN;
								break;
							case 0x25: // time$second
								date_time.second = Z80_DATA_PIN;
								break;
							case 0x0B: // serial$settings
								switch (Z80_DATA_PIN & 0x0F) {
									case 1: serial_init(50); break;
									case 2: serial_init(75); break;
									case 3: serial_init(110); break;
									case 4: serial_init(134); break; // 134.5
									case 5: serial_init(150); break;
									case 6: serial_init(300); break;
									case 7: serial_init(600); break;
									case 8: serial_init(1200); break;
									case 9: serial_init(1800); break;
									case 10: serial_init(2400); break;
									case 11: serial_init(3600); break;
									case 12: serial_init(4800); break;
									case 13: serial_init(7200); break;
									case 14: serial_init(9600); break;
									case 15: serial_init(19200); break;
								}
								if (Z80_DATA_PIN & _BV(4)) {
									serial_handshake = SERIAL_HANDSHAKE_RTS_CTS;
								} else {
									serial_handshake = SERIAL_HANDSHAKE_NONE;
								}
								break;
							case 0x0C: // parallel$base+0
								parallel_write(PARALLEL_DATA, Z80_DATA_PIN);
								break;
							case 0x0D: // parallel$base+1
								parallel_write(PARALLEL_STATUS, Z80_DATA_PIN);
								break;
							case 0x0E: // parallel$base+2
								parallel_write(PARALLEL_CONTROL, Z80_DATA_PIN);
								break;
							case 0x30: // interrupt$enable
								interrupt_enable = Z80_DATA_PIN;
								interrupt_flags &= interrupt_enable;
								if (interrupt_flags) {
									Z80_INT_PORT &= (uint8_t)~_BV(Z80_INT);
								} else {
									Z80_INT_PORT |= _BV(Z80_INT);
								}
								break;
							case 0x31: // interrupt$flags
								interrupt_flags &= (uint8_t)~Z80_DATA_PIN;
								if (interrupt_flags) {
									Z80_INT_PORT &= (uint8_t)~_BV(Z80_INT);
								} else {
									Z80_INT_PORT |= _BV(Z80_INT);
								}
								break;
							case 0x38: // vgm
								vgm_write(Z80_DATA_PIN);
								break;
							default:
								console_output_string_P(PSTR("Output port "));
								console_output_hex_byte(address);
								console_output_string_P(PSTR(" not handled.\r\n"));
								for(;;);
						}
					} else if (!(Z80_RD_PIN & _BV(Z80_RD))) {
						// Respond to Z80 read.
						uint8_t response = 0xFF;
						switch (address) {						
							case 0x00: // vdu
								break;
							case 0x04: // key$status
								response = keyboard_buffer_count > 0 ? 0xFF : 0x00;
								break;
							case 0x05: // key
								response = keyboard_buffer_dequeue();
								break;
							case 0x06: // key$scan
								if (keyboard_key_state[inkey_code / 8] & _BV(inkey_code & 7)) {
									response = 0xFF;
								} else {
									response = 0x00;
								}
								break;
							case 0x08: // serial$out
								response = (UCSR0A & _BV(UDRE0)) ? 0xFF : 0x00;
								break;
							case 0x09: // serial$in$status
								response = (serial_rx_buffer_count > 0) ? 0xFF : 0x00;
								break;
							case 0x0A: // serial$in
								response = serial_receive();
								break;
							case 0x0C: // parallel$base+0
								response = parallel_read(PARALLEL_DATA);
								break;
							case 0x0D: // parallel$base+1
								response = parallel_read(PARALLEL_STATUS);
								break;
							case 0x0E: // parallel$base+2
								response = parallel_read(PARALLEL_CONTROL);
								break;
							case 0x10: // disk$read
								{
									Z80_DATA_PORT = 0;
									Z80_DATA_DDR = 0xFF;
									z80_reset_wait_control_bus();
									wait_reset = 1;
									
									uint32_t flash_offset = (selected_track * (unsigned long)current_partition->records_per_track * 128UL) + (selected_sector * 512UL) + current_partition->offset;
									sd_read_block_notify_errors(flash_offset, sd_block);
									
									z80_memory_control_address();
									for (uint16_t b = 0; b < 512; ++b) {
										z80_memory_set_address(selected_dma + b);
										z80_memory_write(sd_block[b]);
									}
									z80_memory_release_address();
									
									z80_release_bus();
								}
								break;
							case 0x11: // disk$write
								{
									Z80_DATA_PORT = 0;
									Z80_DATA_DDR = 0xFF;
									z80_reset_wait_control_bus();
									wait_reset = 1;
									
									z80_memory_control_address();
									
									for (uint16_t b = 0; b < 512; ++b) {
										z80_memory_set_address(selected_dma + b);
										sd_block[b] = z80_memory_read();
									}
									
									z80_memory_release_address();
									
									uint32_t flash_offset = (selected_track * (unsigned long)current_partition->records_per_track * 128UL) + (selected_sector * 512UL) + current_partition->offset;
									sd_write_block(flash_offset, sd_block);
									
									z80_release_bus();	
								}
								break;
							case 0x20: // time
								date_time_get_rtc(&date_time);
								break;
							case 0x21: // time$date$l
								response = date_time.date;
								break;
							case 0x22: // time$date$h
								response = date_time.date >> 8;
								break;
							case 0x23: // time$hour
								response = date_time.hour;
								break;
							case 0x24: // time$minute
								response = date_time.minute;
								break;
							case 0x25: // time$second
								response = date_time.second;
								break;
							case 0x30: // interrupt$enable
								response = interrupt_enable;
								break;
							case 0x31: // interrupt$flags
								response = ~interrupt_flags;
								break;
							case 0x38: // vgm
								response = vgm_read();
								break;
							default:
								console_output_string_P(PSTR("Input port "));
								console_output_hex_byte(address);
								console_output_string_P(PSTR(" not handled.\r\n"));
								for(;;);
						}
						if (!wait_reset) {
							Z80_DATA_PORT = response;
							Z80_DATA_DDR = 0xFF;
						}
					}
					if (!wait_reset) z80_reset_wait();
				
				} else {
					Z80_DATA_PORT = 0x00;
					Z80_DATA_DDR = 0xFF;
					z80_reset_wait();
				}
			}
		}
		
		console_output_string_P(PSTR("\r\nRestarting...\r\n\r\n"));
	
	}
}