#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>

#include "serial.h"

volatile uint8_t serial_rx_buffer[SERIAL_RX_BUFFER_SIZE];

volatile uint8_t serial_rx_buffer_read_ptr;
volatile uint8_t serial_rx_buffer_write_ptr;
volatile uint8_t serial_rx_buffer_count;

// Defines the current handshake mode.
enum serial_handshake_t serial_handshake;

// Initialises the serial port at a particular baud rate.
void serial_init(uint16_t baud_rate) {
	// Disable handshaking.
	serial_handshake = SERIAL_HANDSHAKE_NONE;
	// Make CTS an input.
	SERIAL_CTS_DDR &= (uint8_t)~_BV(SERIAL_CTS);
	SERIAL_CTS_PORT &= (uint8_t)~_BV(SERIAL_CTS);
	// Make RTS an output and assert it.
	SERIAL_RTS_DDR |= _BV(SERIAL_RTS);
	SERIAL_RTS_PORT &= (uint8_t)~_BV(SERIAL_RTS);
	// Set rate.
	UBRR0 = (F_CPU / 16 / baud_rate) - 1;
	// Enable transmitter, receiver and receiver interrupts.
	UCSR0B = _BV(TXEN0) | _BV(RXEN0) | _BV(RXCIE0);
	// 8 data bits.
	UCSR0C = /*_BV(USBS0) |*/ _BV(UCSZ01) | _BV(UCSZ00);
	// Perform dummy read.
	UDR0 = UDR0;
	serial_received_flag = 0;
	// Empty receive buffer.
	serial_rx_buffer_count = 0;
	serial_rx_buffer_read_ptr = 0;
	serial_rx_buffer_write_ptr = 0;
}

// Determines whether the serial transmitter is ready to send data or not.
uint8_t serial_transmitter_is_ready(void) {
	switch (serial_handshake) {
		case SERIAL_HANDSHAKE_NONE:
			return UCSR0A & _BV(UDRE0);
		case SERIAL_HANDSHAKE_XON_XOFF:
			// TODO: Not implemented.
			return UCSR0A & _BV(UDRE0);
		case SERIAL_HANDSHAKE_RTS_CTS:
			return (UCSR0A & _BV(UDRE0)) && !(SERIAL_CTS_PIN & _BV(SERIAL_CTS));
	}
	return 0;
}

// Transmits a byte out of the serial port. This will block until the transmitter is ready.
void serial_transmit(uint8_t data) {
	// Wait for the data register to become available.
	while (!serial_transmitter_is_ready());
	// Write data.
	UDR0 = data;
}

// Flag set when serial data has been received.
volatile uint8_t serial_received_flag = 0;

ISR(USART0_RX_vect) {
	if (serial_rx_buffer_count < SERIAL_RX_BUFFER_SIZE) {
		++serial_rx_buffer_count;
		serial_rx_buffer[serial_rx_buffer_write_ptr++] = UDR0;
		serial_rx_buffer_write_ptr %= SERIAL_RX_BUFFER_SIZE;
		if (serial_rx_buffer_count >= SERIAL_RX_BUFFER_FULL_THRESHOLD) {
			switch (serial_handshake) {
				case SERIAL_HANDSHAKE_NONE:
					break;
				case SERIAL_HANDSHAKE_XON_XOFF:
					serial_transmit(0x13); // XOFF=^S
					break;
				case SERIAL_HANDSHAKE_RTS_CTS:
					SERIAL_RTS_PORT |= _BV(SERIAL_RTS);
					break;
			}
		}
		serial_received_flag = !0;
	} else {
		// Dummy read and discard the byte.
		UDR0 = UDR0;
	}
}

// Read a byte from the serial port.
uint8_t serial_receive() {
	while (!serial_rx_buffer_count);
	uint8_t result = serial_rx_buffer[serial_rx_buffer_read_ptr++];
	serial_rx_buffer_read_ptr %= SERIAL_RX_BUFFER_SIZE;
	--serial_rx_buffer_count;
	if (serial_rx_buffer_count < SERIAL_RX_BUFFER_FULL_THRESHOLD) {
		switch (serial_handshake) {
			case SERIAL_HANDSHAKE_NONE:
				break;
			case SERIAL_HANDSHAKE_XON_XOFF:
				serial_transmit(0x11); // XON=^Q
				break;
			case SERIAL_HANDSHAKE_RTS_CTS:
				SERIAL_RTS_PORT &= (uint8_t)~_BV(SERIAL_RTS);
				break;
		}
	}
	return result;
}